import collections.abc
import dataclasses
from typing import Any


@dataclasses.dataclass
class EnvValue:
    value: Any
    update: bool = False
    confirmed: bool = False


def attrs_eq(a, b, *args):
    r"""Compare two objects according to their attributes.

    :param a: First object to compare.
    :param b: Second object to compare.
    :param \*args: Attributes names to check for.
    :type \*args: :class:`str`
    :return: `True` if all provided attributes of objects `a` and `b` are equals.
    :rtype: :class:`bool`
    """
    try:
        return all(getattr(a, attr) == getattr(b, attr) for attr in args)
    except AttributeError:
        return False


def attrs_hash(a, *args, **kwargs):
    return hash(tuple(getattr(a, attr) for attr in args) + tuple(kwargs.items()))


def is_nonstring_sequence(obj):
    """Check if a given object is a non-string sequence.

    :param obj: Any object to check.
    :return: `True` if `obj` is a non-string :class:`Sequence` instance.
    :rtype: :class:`bool`
    """
    return (
        isinstance(obj, collections.abc.Sequence)
        and not isinstance(obj, collections.abc.ByteString)
        and not isinstance(obj, str)
    )


def make_repr(obj, *args, **kwargs):
    r"""Make a representation string with ease.

    :param obj: Any object to make a representation for.
    :param \*args: Attributes names to include.
    :type \*args: :class:`str`
    :param \**kwargs: Other `attribute_name=any_value` to include.
    :return: String representation of `obj`.
    :rtype: :class:`str`
    """
    from_attributes = [f"{a}={getattr(obj, a)}" for a in args]
    from_kwargs = [f"{k}={v}" for k, v in kwargs.items()]
    params = ", ".join(from_attributes + from_kwargs)

    return f"""{type(obj).__name__}({params})"""


def unfold_dict(config):
    unfolded_config = {}

    for key_section, value_section in config.items():
        if isinstance(value_section, collections.abc.Mapping):
            if key_section not in unfolded_config:
                unfolded_config[key_section] = {}
            for key, value in value_section.items():
                if isinstance(value, collections.abc.Mapping):
                    unfolded_config[key_section][key] = "\n" + "\n".join(
                        f"{k} = {v}" for k, v in value.items()
                    )
                else:
                    unfolded_config[key_section][key] = value
        else:
            unfolded_config[key_section] = value_section

    return unfolded_config


def unfold_list(config):
    unfolded_config = {}

    for key, value in config.items():
        if isinstance(value, collections.abc.Mapping):
            unfolded_config[key] = unfold_list(value)
        elif is_nonstring_sequence(value):
            unfolded_config[key] = "\n" + "\n".join(value)
        else:
            unfolded_config[key] = value

    return unfolded_config



from abc import ABC


class Tool(ABC):
    """Base Class for any `Tool`. Concrete `Tool` should subclass it."""

    def add_to_structure(self):
        """Add this Tool's configuration and template to project structure`."""

    def __repr__(self):
        return self.__class__.__name__

    def pre(self, workon):
        """Pre-script, done after creating all folders.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """

    def post(self, workon):
        """Post-script, done before writting configuration and template files.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """



import io

from incipyt._internal.utils import is_nonstring_sequence


def mock_stdin(monkeypatch, inputs):
    if not is_nonstring_sequence(inputs):
        inputs = (inputs,)
    monkeypatch.setattr("sys.stdin", io.StringIO("".join(f"{v}\n" for v in inputs)))



from incipyt.tools.base import Tool
from incipyt.tools.git import Git
from incipyt.tools.license import License
from incipyt.tools.pep517 import PDM, Flit, Hatch, Setuptools
from incipyt.tools.poetry import Poetry
from incipyt.tools.venv import Venv



import configparser
import os
import pathlib
from abc import ABC, abstractmethod

import toml

from incipyt._internal import templates, utils


class BaseDumper(ABC):
    def __init__(self, path, sanitizer=None):
        self._path = pathlib.Path(path)
        self._root = None
        self._sanitizer = sanitizer

    @abstractmethod
    def dump_in(self, config):
        raise NotImplementedError

    def commit(self, root):
        self._root = root

        path = self.format_path()
        if path.exists():
            raise FileExistsError(f"File {path} already exists.")

    def mkdir(self):
        self.format_path().parent.mkdir(parents=True, exist_ok=True)

    def open(self, mode="w+", **kwargs):  # noqa: A003
        return self.format_path().open(mode=mode, **kwargs)

    def format_path(self):
        if self._root is None:
            raise RuntimeError("Root is missing.")

        return pathlib.Path(
            templates.FormatterEnviron(sanitizer=self._sanitizer).format(
                os.fspath(self._root / self._path)
            )
        )

    def __hash__(self):
        return hash(self._path)

    def __repr__(self):
        return utils.make_repr(
            self,
            root=self._root,
            path=self._path,
            sanitizer=self._sanitizer,
        )

    def __eq__(self, other):
        return utils.attrs_eq(self, other, "_path")


class CfgIni(BaseDumper):
    def dump_in(self, config):
        config_cfg = configparser.ConfigParser()
        config_cfg.read_dict(utils.unfold_dict(utils.unfold_list(config)))
        with self.open() as file:
            config_cfg.write(file)


class TextFile(BaseDumper):
    def __init__(self, path, sep="\n", sanitizer=None):
        super().__init__(path, sanitizer)
        self._sep = sep

    def dump_in(self, config):
        with self.open() as file:
            file.write(self._sep.join(config) + self._sep)


class Toml(BaseDumper):
    def dump_in(self, config):
        with self.open() as file:
            toml.dump(config, file)



import os
import pathlib
import urllib


def package(key, value):
    return value.replace("-", "_") if key == "NAME" else value


def project(key, value):
    return value.replace("_", "-") if key == "NAME" else value


def url(key, value):
    parts = urllib.parse.urlparse(value)
    parts._replace(path=os.fspath(pathlib.Path(parts.path)))
    return parts.geturl()


def version(key, value):
    return value



from io import StringIO

from pytest import mark

from incipyt._internal.utils import is_nonstring_sequence


@mark.parametrize(
    "obj, res",
    (
        (None, False),
        (b"abc", False),
        ("abc", False),
        (StringIO("abc"), False),
        ({}, False),
        ([], True),
        ((), True),
    ),
)
def test_nonstring_sequence(obj, res):
    assert is_nonstring_sequence(obj) == res



from incipyt.tools.pep517.base import BuildSystem
from incipyt.tools.pep517.flit import Flit
from incipyt.tools.pep517.hatch import Hatch
from incipyt.tools.pep517.pdm import PDM
from incipyt.tools.pep517.setuptools import Setuptools



import logging
import os
import shutil
import sys

import click

from incipyt import project, signals, tools
from incipyt._internal.dumpers import TextFile
from incipyt.commands import git, git_get_config

logger = logging.getLogger(__name__)


class Git(tools.Tool):
    """Scripts to add Git to :class:`incipyt.project._Structure`."""

    def __init__(self):
        if not shutil.which("git"):
            logger.error("%r is missing from the PATH. Abort.", "git")
            sys.exit(1)

        signals.vcs_ignore.connect(self._slot)

    def add_to_structure(self):
        """Add git configuration to `project.structure`.

        Register git related project URLs.

        URLs:
            - Repository: {REPOSITORY}
            - Issue: {REPOSITORY}/issues
            - Documentation: {REPOSITORY}/wiki
        """
        project.structure.use_template("python.gitignore", ".gitignore")
        signals.project_url.emit(url_kind="Documentation", url_value="{REPOSITORY}/wiki")
        signals.project_url.emit(url_kind="Issue", url_value="{REPOSITORY}/issues")
        signals.project_url.emit(url_kind="Repository", url_value="{REPOSITORY}")

    def _slot(self, pattern, **kwargs):
        project.structure.get_config_list(TextFile(".gitignore")).append(pattern)

    def pre(self, workon):
        """Run `git init`.

        Also push {AUTHOR_NAME} and {AUTHOR_EMAIL} using `git config user.*`.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        git(["init", os.fspath(workon)])
        project.environ.setdefault("AUTHOR_EMAIL", git_get_config("user.email", workon=workon))
        project.environ.setdefault("AUTHOR_NAME", git_get_config("user.name", workon=workon))

    def post(self, workon):
        """Check config name+email and then run `git add --all`.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        env_name = project.environ["AUTHOR_NAME"]
        git_name = git_get_config("user.name", workon=workon)
        prompt = f"Git user name is {git_name!r}, use {env_name!r} instead?"
        if env_name != git_name and click.confirm(prompt, default=None):
            git(["config", "user.name", env_name], workon=workon)

        env_email = project.environ["AUTHOR_EMAIL"]
        git_email = git_get_config("user.email", workon=workon)
        prompt = f"Git user email is {git_email!r}, use {env_email!r} instead?"
        if env_email != git_email and click.confirm(prompt, default=None):
            git(["config", "user.email", env_email], workon=workon)

        git(["add", "--all"], workon=workon)



import os
import sys

from incipyt import commands, project, signals, tools
from incipyt._internal import sanitizers, templates
from incipyt._internal.dumpers import Toml

# as of may 2022, the latest stable release of most bsd/linux
# distributions ship a python whoose verion is at least 3.9
LINUX_MIN_PYTHON_VERSION = (3, 9)


class BuildSystem(tools.Tool):
    """Scripts to add generic PEP 517 build system to :class:`incipyt.project._Structure`."""

    def __init__(self, check_build=False, **kwargs):
        self.check_build = check_build
        signals.build_dependency.connect(self._slot_dependency)
        signals.classifier.connect(self._slot_classifier)
        signals.project_url.connect(self._slot_url)

    def add_to_structure(self):
        """Add PEP 517 core structure to `project.structure`.

        :file:`pyptoject.toml`

        .. code-block::
            [project]
            authors = [
                {name = "{AUTHOR_NAME}", email = "{AUTHOR_EMAIL}"}
            ]
            classifiers = [
                "Programming Language :: Python :: 3 :: Only"
            ]
            description = "{SUMMARY_DESCRIPTION}"
            maintainers = [
                {name = "{AUTHOR_NAME}", email = "{AUTHOR_EMAIL}"}
            ]
            name = "{PROJECT_NAME}"
            readme = "README.md"
            requires-python = ">={PYTHON_VERSION}"
            version = "{PACKAGE_VERSION}"

            [project.optional-dependencies]
            dev = [
                "build>=0.2.0",
            ]

        If this configuration cannot be populate like that, an error is raised.

        Here key-value association is appended, if a key already exists the
        value is appended to the current one(s), the user will be asked to
        choose when commiting.
        """
        pyproject = project.structure.get_config_dict(Toml("pyproject.toml"))

        pyproject["project"] = {
            "authors": [{"name": "{AUTHOR_NAME}", "email": "{AUTHOR_EMAIL}"}],
            "description": templates.StringTemplate(
                "{SUMMARY_DESCRIPTION}",
                SUMMARY_DESCRIPTION="\t",
            ),
            "license": {"file": "LICENSE"},
            "maintainers": [{"name": "{AUTHOR_NAME}", "email": "{AUTHOR_EMAIL}"}],
            "name": templates.StringTemplate(
                "{PROJECT_NAME}",
                sanitizer=sanitizers.project,
            ),
            "readme": "README.md",
            "requires-python": templates.StringTemplate(
                ">={AUDIENCE_PYTHON_VERSION}",
                sanitizer=sanitizers.version,
                AUDIENCE_PYTHON_VERSION="{0[0]}.{0[1]}".format(
                    min(sys.version_info, LINUX_MIN_PYTHON_VERSION)
                ),
            ),
            "version": templates.StringTemplate(
                "{PACKAGE_VERSION}",
                sanitizer=sanitizers.version,
                PACKAGE_VERSION="0.0.0",
            ),
        }

        project.structure.use_template("{PROJECT_NAME}/__init__.py", sanitizer=sanitizers.package)
        project.structure.use_template("README.md")

        signals.build_dependency.emit(dep_name="build>=0.2.0")

        signals.classifier.emit(classifier="Programming Language :: Python :: 3 :: Only")

        signals.vcs_ignore.emit(pattern="dist/")
        signals.vcs_ignore.emit(pattern="*.egg-info")

    def _slot_classifier(self, classifier, **kwargs):
        pyproject = project.structure.get_config_dict(Toml("pyproject.toml"))
        if ("project", "classifiers") not in pyproject:
            pyproject["project", "classifiers"] = []
        pyproject["project", "classifiers"].append(classifier)

    def _slot_dependency(self, dep_name, **kwargs):
        pyproject = project.structure.get_config_dict(Toml("pyproject.toml"))
        if ("project", "optional-dependencies", "dev") not in pyproject:
            pyproject["project", "optional-dependencies", "dev"] = []
        pyproject["project", "optional-dependencies", "dev"].append(dep_name)

    def _slot_url(self, url_kind, url_value, **kwargs):
        project.structure.get_config_dict(Toml("pyproject.toml"))["project", "urls"] = {
            url_kind: url_value
        }

    def post(self, workon):
        """Editable install and build for test.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        commands.pip_install(["--editable", f"{workon}[dev]"])
        if self.check_build:
            commands.build([os.fspath(workon)])



from incipyt import project, signals
from incipyt._internal.dumpers import Toml
from incipyt.tools import pep517


class Hatch(pep517.BuildSystem):
    """Scripts to add Hatch to :class:`incipyt.project._Structure`."""

    def add_to_structure(self):
        """Add hatch configuration to `project.structure`.

        :file:`pyptoject.toml`

        .. code-block::

            [build-system]
            build-backend = "hatchling.build"
            requires = ["hatchling>=1.3.0"]

        If this configuration cannot be populate like that, an error is raised.

        :raises RuntimeError: If a build-system is already setup in pyproject.toml.
        """
        super().add_to_structure()

        pyproject = project.structure.get_config_dict(Toml("pyproject.toml"))

        if "build-system" in pyproject:
            raise RuntimeError("Build system already registered.")

        pyproject["build-system"] = {
            "build-backend": "hatchling.build",
            "requires": ["hatchling>=1.3.0"],
        }

        signals.build_dependency.emit(dep_name="hatch>=1.2.0")



"""Signals for tools."""
from signalslot.signal import Signal

build_dependency = Signal(args=["dep_name"])
classifier = Signal(args=["classifier"])
project_url = Signal(args=["url_kind", "url_value"])
vcs_ignore = Signal(args=["pattern"])



"""This module contains classes related to template definition and rendering.

Incipyt mainly uses two kinds of templates: bare template strings that behave
like builtin Python string formatting. It also provides special wrappers to
allow easier templating of collections data structures, such as dict-like
template objects.
"""

import contextlib
import logging
from abc import ABCMeta, abstractmethod
from collections import abc
from pathlib import Path
from string import Formatter

import click

import incipyt
from incipyt import project
from incipyt._internal import utils

logger = logging.getLogger(__name__)


class Formattable(metaclass=ABCMeta):
    @abstractmethod
    def format(self):  # noqa: A003
        raise NotImplementedError


class StringTemplate(Formattable):
    """This class acts like a wrapper around a format string.

    When an instance is called, it renders the underlying format string using
    environ values and overrides.
    """

    def __init__(
        self, format_string, confirmed=False, sanitizer=None, value_error=True, **kwargs
    ):
        r"""This class acts like a wrapper around a format string.

        When an instance is called, it renders the underlying format string
        using environ values and overrides.

        :param format_string: A format string whose keyword argument will be substituted.
        :type format_string: :class:`str`
        :param confirmed: Confirmed status for new variables from keyword args.
        :type confirmed: :class:`bool`, optionnal
        :param sanitizer: An optionnal callable to sanitize the values given (key, value) pairs.
        :type sanitizer: :class:`function` or `None`, optionnal
        :param value_error: Empty values generate errors.
        :type value_error: :class:`bool`
        :param \**kwargs: Variables overrides that will be used for rendering and pushed to the environ.
            Automatically wrapped in :class:`incipyt._internal.utils.EnvValue` if needed.
        :type \**kwargs: :class:`str`, optionnal
        """
        self._confirmed = confirmed
        self._sanitizer = sanitizer
        self._value_error = value_error
        self._format_string = format_string
        self._kwargs = kwargs

    def __eq__(self, other):
        if isinstance(other, str):
            return self.format() == other
        return utils.attrs_eq(
            self, other, "_format_string", "_confirmed", "_sanitizer", "_value_error", "_kwargs"
        )

    def __lt__(self, other):
        if isinstance(other, str):
            return self.format() < other
        return self.format() < other.format()

    def __gt__(self, other):
        if isinstance(other, str):
            return self.format() > other
        return self.format() > other.format()

    def __hash__(self):
        return utils.attrs_hash(
            self, "_format_string", "_confirmed", "_sanitizer", "_value_error", **self._kwargs
        )

    def format(self):  # noqa: A003
        """Format the underlying format string using variables from the environ.

        :return: The formatted string.
        :rtype: :class:`str`
        """
        return FormatterEnviron(sanitizer=self._sanitizer, value_error=self._value_error).format(
            self._format_string,
            **{
                key: (
                    value
                    if isinstance(value, utils.EnvValue)
                    else utils.EnvValue(value, confirmed=self._confirmed)
                )
                for key, value in self._kwargs.items()
            },
        )

    def __repr__(self):
        return utils.make_repr(
            self,
            format_string=self._format_string,
            confirmed=self._confirmed,
            sanitizer=self._sanitizer,
            value_error=self._value_error,
            kwargs=self._kwargs,
        )

    @classmethod
    def wrap(cls, value):
        return value if isinstance(value, Formattable) else cls(value)

    @classmethod
    def from_file(cls, filename):
        path = Path(incipyt.__file__).parent / "templates" / filename
        return cls(path.read_text())


class ChoiceTemplate(Formattable):
    """Class to hold multiple values for a single key.

    When an instance is called, the user will be asked to pick a value using
    the command line interface.
    """

    def __init__(self, head, tail):
        """Class to hold multiple string for a single key.

        When an instance is called, the user will be asked to pick a string
        using the command line interface.

        :param head: Entry to put a the head of the stack.
        :type tail: :class:str
        :param tail: Tail of the stack.
        :type tail: :class:`incipyt._intternal.templates.ChoiceTemplate` or any bare value
        """
        self._values = (
            {StringTemplate.wrap(head)} | tail._values
            if isinstance(tail, ChoiceTemplate)
            else {StringTemplate.wrap(head), StringTemplate.wrap(tail)}
        )

    def format(self):  # noqa: A003
        """Ask the user to pick a value using the command line interface.

        If it is :class:`incipyt._internal.templates.Formattable`, it will be formatted.

        :return: The user-choosen value.
        """
        return click.prompt(
            "Conflicting configuration, choose between",
            type=click.Choice(
                [
                    value.format() if isinstance(value, Formattable) else value
                    for value in self._values
                ]
            ),
        )

    def __eq__(self, other):
        return utils.attrs_eq(self, other, "_values")

    def __hash__(self):
        return hash(tuple(self._values))

    def __repr__(self):
        return f"{type(self).__name__}({self._values})"

    @classmethod
    def from_items(cls, *args):
        r"""Build a class instance from any number of bare values.

        :param \*args: Entries to wrap.
        :return: New class instance
        :rtype: :class:`incipyt._intternal.templates.ChoiceTemplate`
        """
        instance = cls.__new__(cls)
        instance._values = {StringTemplate.wrap(arg) for arg in args}
        return instance


class TemplateDict(abc.MutableMapping):
    """Proxy class around a provided mapping.

    This is itended to ease configuration templating.

    :Class instanciation:

    >>> cfg = TemplateDict({})

    :Setting values:

    Following exemples assume `cfg` is a :class:`incipyt._internal.templates.StringTemplate` around an empty `dict`.

    Bare values will be wrapped into :class:`incipyt._internal.templates.StringTemplate`
    automatically :

    >>> cfg["key"] = "{VARIABLE_NAME}"
    >>> print(cfg)
        TemplateDict(data={'key': StringTemplate(format_string={VARIABLE_NAME})})

    :class:`incipyt._internal.templates.Formattable` will be kept as-is:

    >>> cfg["key"] = a_formattable
    >>> print(cfg)
        TemplateDict(data={'key': a_formattable})

    Collections are supported as well:

    >>> cfg["key"] = ["{VARIABLE_NAME}"]
    >>> print(cfg)
        TemplateDict(data={'key': [StringTemplate("{VARIABLE_NAME}")]})

    >>> cfg["key"] = {"keyB": "{VARIABLE_NAME}"}
    >>> print(cfg)
        TemplateDict(data={'key': {'keyB': StringTemplate(format_string={VARIABLE_NAME})}})

    :Nested keys:

    To ease nested structure definition, the following syntax is supported:

    >>> cfg["keyA", "keyB"] = "{VARIABLE_NAME}"
    >>> print(cfg)
        TemplateDict(data={'keyA': {'keyB': StringTemplate(format_string={VARIABLE_NAME})}})

    :Multiple values:

    Instances of :class:`incipyt._internal.templates.ChoiceTemplate` will be
    created in case of value overrides. For instance, if `previous_value` is a
    :class:`incipyt._internal.templates.Formattable`:

    >>> cfg = TemplateDict({"key": previous_value})
    >>> cfg["key"] = "{VARIABLE_NAME}"
    >>> print(cfg)
        TemplateDict(data={'key': ChoiceTemplate({StringTemplate(format_string={VARIABLE_NAME}), previous_value})})

    If `previous_list` is a mutable sequence, any value not already present in
    it will be appended:

    >>> cfg = TemplateDict({"key": previous_list})
    >>> cfg["key"] = ["{VARIABLE_NAME}"]
    >>> cfg == {"key": previous_list + [StringTemplate("{VARIABLE_NAME}")]}
        True

    :Inplace union:

    The inplace union operator `|=` can be used to set multiple keys at once:

    >>> cfg = TemplateDict({})
    >>> cgf |= {"keyA": "{VARIABLE_NAME}", "keyB": "{OTHER_NAME}"}
    >>> print(cfg)
        TemplateDict(data={'keyA': StringTemplate(format_string={VARIABLE_NAME}), 'keyB': StringTemplate(format_string={OTHER_NAME})})
    """

    def __init__(self, data):
        """Proxy class around a provided mapping.

        This is itended to ease configuration templating.

        :param mapping: Existing mapping holding entries to wrap.
        :type mapping: :class:`collections.abc.MutableMapping`
        """
        self.data = data

    def __getitem__(self, keys):
        if not utils.is_nonstring_sequence(keys):
            keys = (keys,)

        config = self.data

        for key in keys:
            if key not in config:
                raise KeyError(f"Index [{', '.join(keys)}] does not exist.")
            config = config[key]

        if isinstance(config, abc.MutableMapping):
            return TemplateDict(config)
        elif utils.is_nonstring_sequence(config):
            return TemplateList(config)
        else:
            return config

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return utils.make_repr(self, "data")

    def __delitem__(self, key):
        raise NotImplementedError(f"{type(self)} do not support __delitem__, add-only dict-like.")

    def __setitem__(self, keys, value):
        if utils.is_nonstring_sequence(keys):
            config = self.data

            for key in keys[:-1]:
                if key not in config:
                    config[key] = {}
                config = config[key]

            self[keys[:-1]][keys[-1]] = value
            return

        if isinstance(value, abc.Mapping):
            if keys not in self.data:
                self.data[keys] = {}

            if utils.is_nonstring_sequence(self.data[keys]):
                raise TypeError(
                    f"{self.data[keys]}) is already a sequence, cannot set to a dict."
                )
            for k, v in value.items():
                TemplateDict(self.data[keys])[k] = v

        elif utils.is_nonstring_sequence(value):
            if keys not in self.data:
                self.data[keys] = []

            if isinstance(self.data[keys], abc.Mapping):
                raise TypeError(f"{self.data[keys]} is already a mapping, cannot set to a list.")
            TemplateList(self.data[keys]).extend(value)

        else:
            if keys in self.data:
                self.data[keys] = ChoiceTemplate(value, self.data[keys])
            else:
                self.data[keys] = StringTemplate.wrap(value)

    def __ior__(self, other):
        self.update(other)
        return self


class TemplateList(abc.MutableSequence):
    """Proxy class around a provided mapping.

    This is itended to ease configuration templating.

    See :class:`incipyt._internal.templates.StringTemplate` for usage details.
    """

    def __init__(self, data):
        """Proxy class around a provided sequence.

        This is itended to ease configuration templating.

        :param mapping: Existing mapping holding entries to wrap.
        :type mapping: :class:`collections.abc.MutableSequence`
        """
        self.data = data

    def __getitem__(self, index):
        if utils.is_nonstring_sequence(self.data[index]):
            return TemplateList(self.data[index])
        elif isinstance(self.data[index], abc.MutableMapping):
            return TemplateDict(self.data[index])
        else:
            return self.data[index]

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return utils.make_repr(self, "data")

    def __eq__(self, other):
        return utils.attrs_eq(self, other, "data")

    def __setitem__(self, index, value):
        raise NotImplementedError(f"{type(self)} do not support __setitem__, add-only list-like.")

    def __delitem__(self, value):
        raise NotImplementedError(f"{type(self)} do not support del, add-only list-like.")

    def insert(self, index, value):
        if utils.is_nonstring_sequence(value):
            self.data.insert(index, [])
            TemplateList(self.data[index]).extend(value)
        elif isinstance(value, abc.Mapping):
            self.data.insert(index, {})
            TemplateDict(self.data[index]).update(value)
        else:
            new_value = StringTemplate.wrap(value)
            if new_value not in self.data:
                self.data.insert(index, new_value)


class FormatterEnviron(abc.Mapping):
    """Class wrapping an environ and providing an interface to render templates.

    It can be used to render template strings.
    """

    def __init__(self, sanitizer=None, value_error=True):
        """Class wrapping an environ and providing an interface to render templates.

        :param sanitizer: An optionnal callable to sanitize the values given (key, value) pairs.
        :type sanitizer: :class:`function` or `None`, optionnal
        :param environ: Consider empty string value as an error.
        :type environ: :class:`bool`, optional
        """
        self.data = project.environ
        self._keys = set()
        self._sanitizer = sanitizer
        self._value_error = value_error

    def __contains__(self, key):
        if key not in self.data:
            self.data.__getitem__(key)

        return True

    def __getitem__(self, key):
        # Call to inner dict __getitem__ will create missing keys
        value = self.data[key]
        if self._value_error and not value:
            raise ValueError

        return self._sanitizer(key, value) if self._sanitizer else value

    def __iter__(self):
        return iter(self._keys)

    def __len__(self):
        return len(self._keys)

    def keys(self):
        return iter(self)

    def values(self):
        return (self[key] for key in self)

    def items(self):
        return zip(self.keys(), self.values())

    def format(self, format_string, **kwargs):  # noqa: A003
        r"""Render a format string.

        Variables will be request from the underlying environ, and undefined
        variables will be created. If `self._value_error` is `True` and an
        empty variable will cause the whole render result to be `None`.

        Additional variables can be specified using keyword arguments. They
        will be added to the environ, hence they will override environ values.

        :param template: A format string whose keyword argument will be substituted.
        :type template: :class:`str`
        :param \**kwargs: Additional variables that will override environ variables.
        :type \**kwargs: :class:`str`, optionnal
        :return: The rendered template or `None` if a context variable is empty.
        :rtype: :class:`str` or `None`
        """

        self._keys = {item[1] for item in Formatter().parse(format_string) if item[1]}
        for key in self:
            if key in kwargs:
                self.data[key] = kwargs[key]

        with contextlib.suppress(ValueError):
            return format_string.format(**self)



from incipyt import project, signals
from incipyt._internal.dumpers import Toml
from incipyt.tools import pep517


class Flit(pep517.BuildSystem):
    """Scripts to add Flit to :class:`incipyt.project._Structure`."""

    def add_to_structure(self):
        """Add flit configuration to `project.structure`.

        :file:`pyptoject.toml`

        .. code-block::

            [build-system]
            build-backend: "flit_core.buildapi"
            requires: ["flit_core>=3.4.0"]

        If this configuration cannot be populate like that, an error is raised.

        :raises RuntimeError: If a build-system is already setup in pyproject.toml.
        """
        super().add_to_structure()

        pyproject = project.structure.get_config_dict(Toml("pyproject.toml"))

        if "build-system" in pyproject:
            raise RuntimeError("Build system already registered.")

        pyproject["build-system"] = {
            "build-backend": "flit_core.buildapi",
            "requires": ["flit_core>=3.4.0"],
        }

        signals.build_dependency.emit(dep_name="flit>=3.4.0")



from incipyt import project, signals, tools

classifiers = {
    "Copyright": "License :: Other/Proprietary License",
    "Apache 2.0": "License :: OSI Approved :: Apache Software License",
    "BSD 2-Clause": "License :: OSI Approved :: BSD License",
    "BSD 3-Clause": "License :: OSI Approved :: BSD License",
    "CDDL 1.0": "License :: OSI Approved :: Common Development and Distribution License 1.0 (CDDL-1.0)",
    "EPL 2.0": "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
    "GPL 2.0": "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    "GPL 3.0": "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "LGPL 2.0": "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
    "LGPL 2.1": "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
    "LGPL 3.0": "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "MIT": "License :: OSI Approved :: MIT License",
    "MPL 2.0": "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
}


class License(tools.Tool):
    """Include a license to :class:`incipyt.project._Structure`."""

    def add_to_structure(self):
        """Add license file to `project.structure` based on `LICENSE` environ."""
        template_name = "licenses/%s.txt" % project.environ["LICENSE"].replace(" ", "-")
        project.structure.use_template(template_name, dest="LICENSE")

    def pre(self, workon):
        """Emit LICENSE to the classifiers.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        signals.classifier.emit(classifier=classifiers[project.environ["LICENSE"]])
